/*
Very simple GLUT program using bare OpenGL libraries to
demonstrate use of GL_ARB_shading_language_100.

Orion Sky Lawlor, olawlor@acm.org, 2006/10/05 (Public Domain)
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include <GL/glew.h> /* GL Extensions Wrangler, http://glew.sourceforge.net/ */
#include <GL/glut.h> /* GL Utilities Toolkit, http://freeglut.sourceforge.net/ */
#include <GL/glui.h> /* GL User Interface, http://glui.sourceforge.net/ */
#include "osl/vec4.h" /* GLSL-style C++ types, by Orion Lawlor */
#include "osl/mat4.h"
using osl::mat4;

#include "ogl/program.h"
#include "ogl/program.cpp"
#include "ogl/util.cpp"
#include "ogl/fast_mipmaps.c"

#include "ltga.cpp"

#ifndef M_PI
# define M_PI           3.14159265358979323846  /* pi  (mmmmm...) */
#endif


/* Variables updated by the GUI */
float spin_x, spin_y, zoom=1.0, specExp=100.0;
int win_w, win_h; /**< Width and height (in pixels) of our window at last resize */

/* These are rotation matrices, updated by the GLUI_Rotation spheres */
mat4 modelMat(1.0), viewMat(1.0), light1Mat(1.0);


int use_vertex_buffers=1;

float fat=0.0; /* fattening factor--move points along their normals */

/* The thing to draw (sequence of glVertex calls) */
int model_mode=0; /* updated by listbox, below */
const char *model_mode_names[]={
	"File", "Sphere", "Torus", "9 Points", "Triangle Stack",0
};

void Exit(const char *where,const char *why) {
	fprintf (stderr, "FATALITY: OpenGL Error in %s: %s\n", where, why);
	abort();	
}

// Check GL for errors after the last command
void Check(const char *where) {
	GLenum e=glGetError();
	if (e==GL_NO_ERROR) return;
	const GLubyte *errString = gluErrorString(e);
	Exit(where, (const char *)errString);
}

/************* Object we'll be drawing ***********/
class mainObject_t {
public:
	mainObject_t();
	void draw(void);
};


/** Set up a 3D model, stored in a vertex buffer */
mainObject_t::mainObject_t()
{
	/* nothing to set up here... */
}

/* 0-based vertex indices of a triangle */
class triangle {
public:
	int vertices[3]; 
	triangle(int a=-1,int b=-1,int c=-1) {vertices[0]=a;vertices[1]=b;vertices[2]=c;}
	int &operator[](int i) {return vertices[i];}
	int  operator[](int i) const {return vertices[i];}
};
/* properties of a vertex */
class vertex {
public:
	vec3 coord;
	vec3 normal;
	vec3 texcoord;
};

/** A mesh consists of a set of vertices connected by triangles */
class mesh {
public:
	/* properties of each vertex */
	std::vector<vertex> vtx;
	
	/* triangles */
	std::vector<triangle> tri;
	
	/* OpenGL buffer handles to vertex and triangle data */
	GLuint h_vtx, h_tri;
	
	mesh(void) {h_vtx=h_tri=0;}
	
	vec3 triNormal(int i) const;
	/* Shrink mesh down to frac of its previous size */
	mesh *simplify(double frac) const;
	/* Subdivide mesh, creating new points */
	mesh *subdivide(void) const;
	
	void computeNormals(void);
	
	/* New vertex-buffer calls */
	void h_make(void);
	void h_draw(void) const;
	
	/* Old loop-and-glVertex calls: */
	void drawVertex(unsigned int i) const;
	void drawTriangle(const triangle &t) const;
	void draw(void) const;
	
};


/* Shrink mesh down to frac of its previous size.
  For the real way to do this, see
  	http://www.uiuc.edu/~garland/software/qslim.html
*/
mesh *mesh::simplify(double frac) const
{
	mesh *nu=new mesh; /* simplified mesh */
	
	/* newFmOld[old vertex number]=new vertex number
	*/
	std::vector<int> newFmOld(vtx.size());
	unsigned int v,f;
	for (v=0;v<vtx.size();v++) newFmOld[v]=v;
	std::vector<vertex> newVtx; /* new vertices */
	
	/* Decimate some triangles */
	int ndecim=(int)(frac*tri.size()/4); /* each decimation removes up to 4 triangles */
	for (int d=0;d<ndecim;d++) 
	{ /* Pick a triangle to decimate.  Note that random choices *don't* work well at all!  */
		int tryNo=0;
	pickDoomed:
		int doomed=(rand()+tryNo++)%tri.size();
		
		/* Check if this triangle is already doomed */
		int a=newFmOld[tri[doomed][0]];
		int b=newFmOld[tri[doomed][1]];
		int c=newFmOld[tri[doomed][2]];
		if (a==b || b==c || c==a) { /* this triangle's already squished */
			goto pickDoomed;
		}
		
		/* Get rid of the doomed triangle, by collapsing its vertices into vertex 0 */
		for (int i=1;i<3;i++) {
			newFmOld[tri[doomed][i]]=newFmOld[tri[doomed][0]];
		}
	}
	
	/* Copy vertex data into new array */
	nu->vtx.resize(vtx.size());
	for (v=0;v<vtx.size();v++) {
		nu->vtx[newFmOld[v]]=vtx[v];
	}
	
	/* Transform old triangles into new vertex numbers */
	for (f=0;f<tri.size();f++) {
		// Look up this triangle's new vertex numbers
		int a=newFmOld[tri[f][0]];
		int b=newFmOld[tri[f][1]];
		int c=newFmOld[tri[f][2]];
		if (a==b || b==c || c==a) continue; /* skip this triangle-- it's been squished */
		nu->tri.push_back(triangle(a,b,c));
	}
	printf("Decimated %d vertices, %d tri into %d vertices, %d tri\n",
		vtx.size(),tri.size(), nu->vtx.size(),nu->tri.size());
	return nu;
}

/* Subdivide mesh, creating new points */
mesh *mesh::subdivide(void) const {
	mesh *nu=new mesh;
	nu->vtx=vtx; /* copy all old vertices */
	/* Simple hack: divide each triangle into four pieces */
	unsigned int v,t;
	unsigned int nt=tri.size();
	for (t=0;t<nt;t++)
	{
		enum {n=3};
		int end[n]; /* old vertex at corner of triangle */
		int mid[n]; /* new vertex along edge of triangle */
		/* Add a vertex along each edge of this triangle */
		for (int i=0;i<n;i++) {
			int prev=tri[t][i], next=tri[t][(i+1)%n];
			end[i]=prev;
			/* Average prev and next to give midpoint vertex. 
			 FIXME: adjacent triangle will want to add same vertex along shared edge; 
			 so should keep a std::map indexed by start and end vertex
			 to prevent duplicate insertions 
			 (this would cut # of inserted vertices by a factor of 2!)
			*/
			vertex m;
			m.coord=0.5*(vtx[prev].coord+vtx[next].coord);
			m.normal=normalize(vtx[prev].normal+vtx[next].normal);
			m.texcoord=0.5*(vtx[prev].texcoord+vtx[next].texcoord);
			
			/* Bias up midpoint, along normal, depending on curvature.
			  This assumes the surface is a positive-curvature sphere;
			  this is *exact* for spheres, but fails miserably in convex
			  regions.  See some other, better algorithms at: 
				http://en.wikipedia.org/wiki/Subdivision_surface
			*/
			double cosAng=dot(vtx[prev].normal,m.normal);
			double sinAng=length(cross(vtx[prev].normal,m.normal));
			double chord=length(vtx[prev].coord-vtx[next].coord);
			double r=0.5*chord/sinAng; /* radius of assumed sphere */
			double bias=r-cosAng*r;
			m.coord+=bias*m.normal;
			
			mid[i]=nu->vtx.size();
			nu->vtx.push_back(m);
		}
		/* Make new subdivided triangles */
		nu->tri.push_back(triangle(end[0],mid[0],mid[2]));
		nu->tri.push_back(triangle(mid[0],end[1],mid[1]));
		nu->tri.push_back(triangle(mid[0],mid[1],mid[2]));
		nu->tri.push_back(triangle(mid[2],mid[1],end[2]));
	}
	
	printf("Subdivided %d vertices, %d tri into %d vertices, %d tri\n",
		vtx.size(),tri.size(), nu->vtx.size(),nu->tri.size());
	return nu;
}



/** Set up vertex and triangle buffers */
void mesh::h_make(void)
{
/* Copy in all the vertex data */
	glGenBuffersARB(1,&h_vtx); 
	glBindBufferARB(GL_ARRAY_BUFFER_ARB,h_vtx);
	glBufferDataARB(GL_ARRAY_BUFFER_ARB,sizeof(vtx[0])*vtx.size(),&vtx[0],
		GL_STATIC_DRAW_ARB);

/* Describe how the vertex data is laid out: these are byte offsets into the vertex struct */
	glEnableClientState(GL_VERTEX_ARRAY);
	glVertexPointer(3,GL_FLOAT,sizeof(vtx[0]),
		(void *)0);
	
	glEnableClientState(GL_NORMAL_ARRAY);
	glNormalPointer(GL_FLOAT,sizeof(vtx[0]),
		(void *)(sizeof(vtx[0].coord)));
	
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glTexCoordPointer(3,GL_FLOAT,sizeof(vtx[0]),
		(void *)(sizeof(vtx[0].coord)+sizeof(vtx[0].normal)));

/* Copy in all the triangle index data (it's zero-based by default) */
	glGenBuffersARB(1,&h_tri);
	glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB,h_tri);
	glBufferDataARB(GL_ELEMENT_ARRAY_BUFFER_ARB,sizeof(tri[0])*tri.size(),&tri[0],
		GL_STATIC_DRAW_ARB); // GL_STREAM_DRAW_ARB
	
/* Switch back to the default buffers */
	glBindBufferARB(GL_ARRAY_BUFFER_ARB,0);
	glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB,0);
}

/** Draw vertices and triangles, by binding vertex and triangle buffers, and calling glDrawElements */
void mesh::h_draw(void) const
{
	glBindBufferARB(GL_ARRAY_BUFFER_ARB,h_vtx);
	glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB,h_tri);
	
	glDrawElements(GL_TRIANGLES,3*tri.size(),GL_UNSIGNED_INT,0);
	
	glBindBufferARB(GL_ARRAY_BUFFER_ARB,0);
	glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB,0);
}

vec3 mesh::triNormal(int i) const {
	const triangle &t=tri[i];
	const vec3 &a=vtx[t[0]].coord;
	const vec3 &b=vtx[t[1]].coord;
	const vec3 &c=vtx[t[2]].coord;
	
	vec3 ba=b-a, ca=c-a;
	vec3 n=normalize(cross(ba,ca));
	
	return n;
}

void mesh::computeNormals(void)
{
	unsigned int v,t;
	/* zero out vertex normals */
	for (v=0;v<vtx.size();v++)
		vtx[v].normal=vec3(0.0);
	
	/* add triangle normals to adjacent vertices */
	for (t=0;t<tri.size();t++) {
		vec3 n=triNormal(t);
		for (int i=0;i<3;i++) {
			int v=tri[t][i];
			vec3 me=vtx[v].coord;
			vec3 a=vtx[tri[t][(i+1)%3]].coord-me;
			vec3 b=vtx[tri[t][(i+2)%3]].coord-me;
			double cosAngle=dot(a,b)/(a.mag()*b.mag());
			double angle=acos(cosAngle);
			double weight=angle; /* weight normal by angle triangle makes */
			vtx[v].normal+=weight*n;
		}
	}
	
	/* normalize vertex normals */
	for (v=0;v<vtx.size();v++)
		vtx[v].normal=normalize(vtx[v].normal);
	
}

void mesh::drawVertex(unsigned int i) const {
	glNormal3fv(vtx[i].normal);
	glTexCoord3fv(vtx[i].texcoord);
	glVertex3fv(vtx[i].coord);
}

void mesh::drawTriangle(const triangle &t) const {
	drawVertex(t[0]);
	drawVertex(t[1]);
	drawVertex(t[2]);
}

void mesh::draw(void) const
{
	glBegin(GL_TRIANGLES);
	for (unsigned int i=0;i<tri.size();i++)
		drawTriangle(tri[i]);
	glEnd();
}

/* Read .smf or .obj file into a mesh object */
const char *readSMF(const char *fName,mesh &m) 
{
	FILE *f=fopen(fName,"r");
	if (f==NULL) return "Couldn't open SMF file!";
	
	double scale=0.0;
	int vt=0; /* texture coordinates vertex counter */
	/* Walk through the lines in the input file */
	enum {max_line=1000};
	char line[max_line];
	while (0!=fgets(line,max_line,f)) {
		char what[max_line];
		what[0]=0; /* empty */
		sscanf(line,"%s",what); /* get the first thing on the line */
		if (0==strcmp(what,"v")) 
		{ /* vertex coordinates */
			double x,y,z;
			if (4!=sscanf(line,"%s%lf%lf%lf",what,&x,&y,&z))
				return "Parse error on 'v' line!";
			// Transform coordinates so object is in a reasonable orientation
			vertex v;
			if (scale==0) scale=1.0/fabs(z); /*<- silly hack: first point's Z sets scale */
			v.coord=scale*vec3(-x,z,y);
			v.texcoord=vec3(0,0,0);
			m.vtx.push_back(v);
		} 
		else if (0==strcmp(what,"vt")) 
		{ /* vertex texture coordinates */
			double x,y,z=0.0;
			if (3<sscanf(line,"%s%lf%lf%lf",what,&x,&y,&z))
				return "Parse error on 'vt' line!";
			m.vtx[vt++].texcoord=vec3(x,y,z);
		} 
		else if (0==strcmp(what,"f")) 
		{ /* face (triangle) */
			int a,b,c,n; /* 1-based vertex numbers */
			if (4!=sscanf(line,"%s%d%d%d",what,&a,&b,&c)) {
				/* also support weird normal-containing syntax */
				if (7!=sscanf(line,"%s%d/%d %d/%d %d/%d", what,&a,&n,&b,&n,&c,&n)) {
					return "Parse error on 'f' line!";
				}
			}
			triangle t;
			t[0]=a-1; 
			t[1]=b-1; 
			t[2]=c-1; 
			m.tri.push_back(t);
		} 
		/* else just skip unknown lines, comments, etc. */
	}
	
	fclose(f);
	return 0;
}


mesh *read_mesh(const char *fName) 
{
	mesh *m=new mesh;
	const char *err=readSMF(fName,*m);
	if (err) {printf("Error reading mesh: %s\n",err); exit(1);}
	m->computeNormals();
	while (m->tri.size()<100000u) { /* subdivide surface mesh */
		mesh *old_m=m;
		m=m->subdivide(); 
		delete old_m;
	}
	//m->computeNormals();
	m->h_make();
	return m;
}

/* Draw some geometry from a file */
void drawFile(void) 
{
	glFinish(); /* avoid timer corruption */
	double start=oglTimer();
	
	static oglTexture *tex=0;
	if (tex==0) {
		LTGA img("thingy.tga");
		if (img.GetImageWidth()!=0) {
			tex=new oglTexture(img.GetPixels(),
				img.GetImageWidth(),img.GetImageHeight(),
				oglTexture_mipmap,
				img.GetImageType()==itRGB?GL_RGB:GL_RGBA);
		}
	}
	if (tex) tex->bind();
	
	static mesh *m=read_mesh("thingy.smf");
	if (use_vertex_buffers)
		m->h_draw();
	else 
		m->draw();
	
	glFinish();
	double elapsed=oglTimer()-start;
	printf("Drawing took %.3f seconds (%d vertex, %d tris)\n",
		elapsed,m->vtx.size(),m->tri.size());
}


/* Sphere drawn as a set of points */
void drawSphere(void)
{
	glColor4f(1,0,1,1.0); /* purple */
	glutSolidSphere(1.0,40,60);
}

/* Draw a torus */
void drawTorus(void) {
	glColor4f(1,1,0,1.0); /* yellow */
	glutSolidTorus(0.2,0.5,40,60);
}

/* Draw a set of points */
void drawPoints(void) 
{
	const static vec4 verts[]={
		vec4(0.0,0,0,1),
		vec4(0.0,1,0,1),
		vec4(0.3,0,0,1),
		vec4(0.3,1,0,1),
		vec4(0.6,0,0,1),
		vec4(0.6,1,0,1),
		vec4(0.9,0,0,1),
		vec4(0.9,1,0,1),
		vec4(1.2,0,0,1),
	};
	glBegin(GL_POINTS);
	glNormal3f(0,0,+1);
	glColor4f(1,0,0,0.5); /* Red points */
	for (unsigned int i=0;i<sizeof(verts)/sizeof(verts[0]);i++) 
		glVertex4fv(verts[i]);
	glEnd();
}

/* Draw a stack of triangles */
void drawStack(void) 
{
	glBegin(GL_TRIANGLES);
	glNormal3f(0,0,+1);
	for (float bz=-0.25; bz<=+0.25;bz+=0.25) {
		glColor4f(1,0,0,1); /* Red triangles */
		glVertex4f(0,0,bz,1);
		glVertex4f(1,0,bz,1);
		
		glColor4f(1,1,0,1); /* Yellow vertex */
		glVertex4f(0,1,bz,1);

		glColor4f(0,1,0,1); /* Green triangles */
		float z=bz+0.1;
		glVertex4f(1,1,z,1);
		glVertex4f(1,0,z,1);
		glVertex4f(0,1,z,1);
	}
	glEnd();
}

/** Transpose this mat4.  If the mat4 was a rotation matrix, 
  this actually inverts the matrix. 
*/
mat4 transpose(const mat4 &m) {
	return mat4(
		m[0][0], m[1][0], m[2][0], m[3][0],
		m[0][1], m[1][1], m[2][1], m[3][1],
		m[0][2], m[1][2], m[2][2], m[3][2],
		m[0][3], m[1][3], m[2][3], m[3][3]);
}


/** Main drawing routine, called by main_display below */
void mainObject_t::draw(void)
{
	oglCheck("start draw");
	/* Set up programmable shader.
	  In a real program, you'd do all your glCreateProgramObjectARB 
	  calls at startup, instead of recompiling the program every frame... */
	static oglProgramObject *p=oglProgramFromFiles("vertex.txt","fragment.txt");
	
	// Everything we draw from here will use our programmable shader...
	p->begin();
	
	/* Copy spin_x and spin_y into program uniforms */
	p->set("fat",fat);
	p->set("tex",0);
	//p->set("spin_y",spin_y);
	p->set("specExp",specExp);
	
	/**
	 Coordinate systems used here:
	
	Model coordinates, which are those passed in to glVertex
		modelMat transforms to world coordinates
	World coordinates, where lighting happens
		viewMat transforms to eye coordinates
	Eye coordinates, where the camera lies on the Z axis
		persMat transforms to clip coordinates
	Clip coordinates, onscreen position from -1 to +1 in X and Y
		the graphics hardware internally transforms clip coordinates to
	Window coordinates, which are just onscreen pixels.
	
	The vertex shader reads model coordinates in gl_Vertex
	and writes clip coordinates in gl_Position.
	*/
	mat4 modelMat=::modelMat; /* read rotation part of model matrix */
	//modelMat[3].x+=0.3; /* shift model slightly in X direction*/
	p->set("modelMat",modelMat);
	p->set("lightDir",light1Mat*vec4(0,0,1,1));
	
	// Figure out where the camera is supposed to be
	double cameraDist=7.0; // Distance of camera from origin, in perspective matrix
	vec4 cameraLoc=vec4(0,0,cameraDist,1.0); // Location of camera after viewMat, and before perspective matrix
	mat4 viewInv=transpose(viewMat); // inverse is transpose for rotation matrix
	cameraLoc=viewInv*cameraLoc; // Location of camera in world coordinates
	p->set("cameraLoc",cameraLoc);
	
	// Perspective matrix
	mat4 persMat=mat4(1.0);
	persMat[0].x=cameraDist*zoom*win_h/(float)win_w; /* aspect ratio correction */
	persMat[1].y=cameraDist*zoom;
	persMat[2].z=-0.1; /* FIXME: set depth range in some smarter way... */
	persMat[2].w=-1.0; /* Flip Z axis, so -Z is deeper into screen (bigger W) */
	persMat[3].w=cameraDist;
	
	// Overall coordinate transform: perspective  view   model
	mat4 clipMat=persMat*viewMat*modelMat; 
	p->set("clipMat",clipMat);
	
	/* Finally, at long last, draw some friggin' objects! */
	switch (model_mode) {
	case 0: drawFile(); break;
	case 1: drawSphere(); break;
	case 2: drawTorus(); break;
	case 3: drawPoints(); break;
	case 4: drawStack(); break;
	};
	
	/* Stop using programmable shader */
	p->end();
	//delete p;
	
	/* This line causes the world to repeatedly redraw itself, using 100% of the CPU. */
	glutPostRedisplay();
	oglCheck("end draw");
}


/******** Program Initialization *********/
mainObject_t *mainObject=0; /**< Big object we're trying to draw */


/* GLUT calls this routine every time our window needs to be redrawn */
void main_display(void) {
	Check("Entering display routine");
	// Clear the color and depth buffers
	glClearColor(0.2,0.3,0.4,1.0); /* set clear color */
	glClearDepth(1.0); /* set clear depth */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); /* actually clear color and depth */
	
	// Set up a little error-prone OpenGL state:
	glDisable(GL_POLYGON_SMOOTH); /* smooth edges of polygons */
	glDisable(GL_BLEND); /* alpha blending */
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_DEPTH_TEST);
	glPointSize(10.0); /* diameter in pixels */
	glLineWidth(10.0); /* thickness in pixels */
	
	// Draw stuff
	Check("About to draw");
	mainObject->draw();
	Check("Done drawing");
	
	// Display the stuff drawn so far to the screen:
	glutSwapBuffers();
}

/* GLUT calls this routine every time our window changes size */
void main_reshape(int w, int h) {
	Check("Before glViewport");
	win_w=w; win_h=h-50; //<- hack: leave room at top for GLUI spinner boxes
	if (win_w<0) win_w=0; // <- Prevents "invalid value" error for small windows...
	if (win_h<0) win_h=0;
	glViewport(0, 0, (GLsizei)win_w, (GLsizei)(win_h));
	Check("After glViewport");
}

void make_list(GLUI *glui,const char *title,int *mode_ptr,const char **names)
{
	GLUI_Listbox *list = new GLUI_Listbox( glui, title, mode_ptr);
	for (int i=0;names[i];i++) 
		list->add_item(i,names[i]);
	list->do_selection(*mode_ptr);
}

int main(int argc, char** argv)
{
	glutInitWindowSize(800,600);
	glutInit(&argc,argv);

	// Ask for a back buffer, color buffer, and Z buffer in a window.
	int mode=GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH;
	glutInitDisplayMode (mode);
	int main_window = glutCreateWindow("GLSL Lighting Demo");
	
	// Turn on OpenGL extensions, like programmable shaders
	//   WARNING: you MUST call glutCreateWindow before doing this!
	//   If you don't call glewInit, or call it wrong, you'll segfault!
	glewInit();

	// Ask GLUT to call us when we need to redraw
	glutDisplayFunc (main_display);
	glutReshapeFunc (main_reshape);
	
	// Make an object to draw
	mainObject=new mainObject_t;
	
	// Make some sliders (this is the only GLUI code you'll need!)
	GLUI *glui = GLUI_Master.create_glui_subwindow(main_window, 
		GLUI_SUBWINDOW_TOP);

	GLUI_Spinner *spinner_x = new GLUI_Spinner( glui, "X:", &spin_x);
	spinner_x->set_float_limits(0.0, 1.0);
/*	GLUI_Spinner *spinner_y = new GLUI_Spinner( glui, "Y:", &spin_y);
	spinner_y->set_float_limits(0.0, 1.0);
*/
	GLUI_Spinner *spin_zoom = new GLUI_Spinner( glui, "Zoom:", &zoom);
	spin_zoom->set_float_limits(0.0, 10.0);
	GLUI_Spinner *spin_spec = new GLUI_Spinner( glui, "Fattening:", &fat);
	spin_spec->set_float_limits(-0.4, +0.4);

	new GLUI_Column(glui);
	
	make_list(glui,"Model:",&model_mode,model_mode_names);
	new GLUI_Checkbox(glui,"Use vertex buffers",&use_vertex_buffers);
	
	new GLUI_Column(glui);
	new GLUI_Rotation(glui,"Model:",&modelMat[0][0]);
	new GLUI_Column(glui,false);
	new GLUI_Rotation(glui,"View:",&viewMat[0][0]);
	new GLUI_Column(glui,false);
	new GLUI_Rotation(glui,"Light:",&light1Mat[0][0]);
	
	// Process events until user closes our window...
	glutMainLoop();
	
	return 0;
}
